searchNodes=[{"doc":"DataSchemas are declarative specifications of how to create structs from some kind of data source. For example you can define a schema that knows how to turn an elixir map into a struct, casting all of the values as it goes. Alternatively you can set up a schema to ingest XML data and create structs from the values inside the XML. Below is an example of a simple schema: defmodule Blog do import DataSchema , only : [ data_schema : 1 ] data_schema ( [ field : { :name , &quot;name&quot; , &amp; to_string / 1 } ] ) end This says we will create a struct with a :name key and will get the value for that key from under the &quot;name&quot; key in the source data. That value will be passed to to_string/1 and the result of that function will end up as the value under :name in the resulting struct. In general this is the format for a field: field { :content , &quot;text&quot; , &amp; cast_string / 1 } # ^ ^ ^ ^ # field type | | | # struct key name | | # path to data in the source | # casting function Field Types There are 4 kinds of struct fields we can have: field - The value will be a casted value from the source data. list_of - The value will be a list of casted values created from the source data. has_one - The value will be created from a nested data schema (so will be a struct) aggregate - The value will a casted value formed from multiple bits of data in the source. Examples See the guides for more in depth examples but below you can see how we create a schema that will take a map of data and create a struct out of it. Given the following schema: defmodule Sandwich do require DataSchema DataSchema . data_schema ( [ field : { :type , &quot;the_type&quot; , &amp; String . upcase / 1 } , list_of : { :fillings , &quot;the_fillings&quot; , &amp; ( String . downcase ( &amp;1 [ &quot;name&quot; ] ) ) } ] ) end input_data = %{ &quot;the_type&quot; =&gt; &quot;fake steak&quot; , &quot;the_fillings&quot; =&gt; [ %{ &quot;name&quot; =&gt; &quot;fake stake&quot; , &quot;good?&quot; =&gt; true } , %{ &quot;name&quot; =&gt; &quot;SAUCE&quot; } , %{ &quot;name&quot; =&gt; &quot;sweetcorn&quot; } , ] } DataSchema . to_struct! ( input_data , Sandwich ) # outputs the following: % Sandwich { type : &quot;FAKE STEAK&quot; , fillings : [ &quot;fake stake&quot; , &quot;sauce&quot; , &quot;sweetcorn&quot; ] , }","ref":"DataSchema.html","title":"DataSchema","type":"module"},{"doc":"Defines a data schema with the provided fields. Uses the default DataSchema.MapAccessor as the accessor, meaning it will expect the source data to be an elixir map and will use Map.get/2 to access the required values in the source data. See DataSchema.data_schema/2 for more details on what fields should look like.","ref":"DataSchema.html#data_schema/1","title":"DataSchema.data_schema/1","type":"macro"},{"doc":"A macro that creates a data schema. By default all struct fields are required but you can specify that a field be optional by passing the correct option in. See the Options section below for more. Field Types There are 4 kinds of struct fields we can have: field - The value will be a casted value from the source data. list_of - The value will be a list of casted values created from the source data. has_one - The value will be created from a nested data schema (so will be a struct) aggregate - The value will a casted value formed from multiple bits of data in the source. Options Available options are: :optional? - specifies whether or not the field in the struct should be included in the @enforce_keys for the struct. By default all fields are required but you can mark them as optional by setting this to true . Examples See the guides for more in depth examples but below you can see how we create a schema that will take a map of data and create a struct out of it. Given the following schema: defmodule Sandwich do require DataSchema DataSchema . data_schema ( [ field : { :type , &quot;the_type&quot; , &amp; String . upcase / 1 } , list_of : { :fillings , &quot;the_fillings&quot; , &amp; ( String . downcase ( &amp;1 [ &quot;name&quot; ] ) ) } ] ) end input_data = %{ &quot;the_type&quot; =&gt; &quot;fake steak&quot; , &quot;the_fillings&quot; =&gt; [ %{ &quot;name&quot; =&gt; &quot;fake stake&quot; , &quot;good?&quot; =&gt; true } , %{ &quot;name&quot; =&gt; &quot;SAUCE&quot; } , %{ &quot;name&quot; =&gt; &quot;sweetcorn&quot; } , ] } DataSchema . to_struct! ( input_data , Sandwich ) # outputs the following: % Sandwich { type : &quot;FAKE STEAK&quot; , fillings : [ &quot;fake stake&quot; , &quot;sauce&quot; , &quot;sweetcorn&quot; ] , }","ref":"DataSchema.html#data_schema/2","title":"DataSchema.data_schema/2","type":"macro"},{"doc":"","ref":"DataSchema.html#to_struct/2","title":"DataSchema.to_struct/2","type":"function"},{"doc":"Accepts an data schema module and some source data and attempts to create the struct defined in the schema from the source data recursively. We essentially visit each field in the schema and extract the data the field points to from the sauce data, passing it to the field's casting function before setting the result of that as the value on the struct. This function takes a simple approach to creating the struct - whatever you return from a casting function will be set as the value of the struct field. You should raise if you want casting to fail. Examples data = %{ &quot;spice&quot; =&gt; &quot;enables space travel&quot; } defmodule Foo do require DataSchema DataSchema . data_schema ( field : { :a_rocket , &quot;spice&quot; , &amp; &amp;1 } ) end DataSchema . to_struct! ( data , Foo ) # =&gt; Outputs the following: % Foo { a_rocket : &quot;enables space travel&quot; }","ref":"DataSchema.html#to_struct!/2","title":"DataSchema.to_struct!/2","type":"function"},{"doc":"Provides helper functions for schemas where the input data is a map.","ref":"DataSchema.Map.html","title":"DataSchema.Map","type":"module"},{"doc":"Allows us to just provide fields to a schema definition by partially applying the DataSchema.Map module. In essence it allows this: defmodule Blog do import DataSchema.Map , only : [ map_schema : 2 ] map_schema ( [ field : { :name , &quot;name&quot; , &amp; to_string / 1 } ] ) end Rather than this: defmodule Blog do import DataSchema , only : [ data_schema : 2 ] data_schema ( [ field : { :name , &quot;name&quot; , &amp; to_string / 1 } ] , MapAccessor ) end Which looks a bit nicer and reduces the surface area of future changes.","ref":"DataSchema.Map.html#map_schema/1","title":"DataSchema.Map.map_schema/1","type":"macro"},{"doc":"Defines a way to use elixir maps as an input data source. This will Map.get/2 the fields from the source data, meaning it will return nil in the case of the key being missing.","ref":"DataSchema.MapAccessor.html","title":"DataSchema.MapAccessor","type":"module"},{"doc":"Accesses the source data using Map.get/2 when an :aggregate type is encountered during struct creation.","ref":"DataSchema.MapAccessor.html#aggregate/2","title":"DataSchema.MapAccessor.aggregate/2","type":"function"},{"doc":"Accesses the source data using Map.get/2 when a :field type is encountered during struct creation.","ref":"DataSchema.MapAccessor.html#field/2","title":"DataSchema.MapAccessor.field/2","type":"function"},{"doc":"Accesses the source data using Map.get/2 when a :has_one type is encountered during struct creation.","ref":"DataSchema.MapAccessor.html#has_one/2","title":"DataSchema.MapAccessor.has_one/2","type":"function"},{"doc":"Accesses the source data using Map.get/2 when a :list_of type is encountered during struct creation.","ref":"DataSchema.MapAccessor.html#list_of/2","title":"DataSchema.MapAccessor.list_of/2","type":"function"},{"doc":"Defines how DataSchema.to_struct!/2 should access data for each given field type. When we create a struct defined by the schema we will visit each struct field in turn and attempt to extract values from the source data. You tell the schema how to extract that data by providing a module that implements this behaviour. Modules that implement this behaviour are implementing how to access data in the source for each kind of field. Examples To define a schema that consumes a map to create a struct you would first create a module that implements this behaviour like so: defmodule DataSchema.MapAccessor do @behaviour DataSchema.DataAccessBehaviour @impl true def field ( data , field ) do Map . get ( data , field ) end @impl true def list_of ( data , field ) do Map . get ( data , field ) end @impl true def has_one ( data , field ) do Map . get ( data , field ) end @impl true def aggregate ( data , field ) do Map . get ( data , field ) end end Then you would pass it into the schema definition: defmodule BlogPost do import DataSchema , only : [ data_schema : 2 ] data_schema ( [ field : { :content , &quot;content&quot; , &amp; to_string / 1 } ] , DataSchema.MapAccessor ) end When we call DataSchema.to_struct!/2 the functions in DataSchema.MapAccessor will be used to access the data from the source. DataSchema ships with the above example (see DataSchema.MapAccessor ) and you can use DataSchema.data_schema/1 to leverage it automatically.","ref":"DataSchema.DataAccessBehaviour.html","title":"DataSchema.DataAccessBehaviour","type":"behaviour"},{"doc":"The function that will be called when an :aggregate field is encountered in the schema when we are turning some input data into a struct.","ref":"DataSchema.DataAccessBehaviour.html#c:aggregate/2","title":"DataSchema.DataAccessBehaviour.aggregate/2","type":"callback"},{"doc":"The function that will be called when a :field field is encountered in the schema when we are turning some input data into a struct.","ref":"DataSchema.DataAccessBehaviour.html#c:field/2","title":"DataSchema.DataAccessBehaviour.field/2","type":"callback"},{"doc":"The function that will be called when a :has_one field is encountered in the schema when we are turning some input data into a struct.","ref":"DataSchema.DataAccessBehaviour.html#c:has_one/2","title":"DataSchema.DataAccessBehaviour.has_one/2","type":"callback"},{"doc":"The function that will be called when a :list_of field is encountered in the schema when we are turning some input data into a struct.","ref":"DataSchema.DataAccessBehaviour.html#c:list_of/2","title":"DataSchema.DataAccessBehaviour.list_of/2","type":"callback"},{"doc":"An error for when a schema is specified incorrectly. Usually this means there has been a syntax error when defining the schema.","ref":"DataSchema.InvalidSchemaError.html","title":"DataSchema.InvalidSchemaError","type":"exception"},{"doc":"Data schemas are declarative descriptions of how to create a struct from some input data. You can set up different schemas to handle different kinds of input data. By default we assume the incoming data is a map, but you can configure schemas to work with any arbitrary data input including XML and json. Data Schemas really shine when working with API responses - converting the response into trusted internal data easily and efficiently.","ref":"getting-started.html","title":"What Are Data Schemas?","type":"extras"},{"doc":"Let's think of creating a struct as taking some source data and turning it into the desired struct. To do this we need to know at least three things: The keys of the desired struct The types of the values for each of the keys Where / how to get the data for each value from the source data. Turning the source data into the correct type defined by the schema will often require casting, so to cater for that the type definitions are casting functions. Let's look at a simple field example field { :content , &quot;text&quot; , &amp; cast_string / 1 } # ^ ^ ^ # struct field | | # path to data in the source | # casting function This says in the source data there will be a field called :text . When creating a struct we should get the data under that field and pass it too cast_string/1 . The result of that function will be put in the resultant struct under the key :content . There are 4 kinds of struct fields we could want: field - The value will be a casted value from the source data. list_of - The value will be a list of casted values created from the source data. has_one - The value will be created from a nested data schema. aggregate - The value will a casted value formed from multiple bits of data in the source. To see this better let's look at a very simple example. Assume our input data looks like this: source_data = %{ &quot;content&quot; =&gt; &quot;This is a blog post&quot; , &quot;comments&quot; =&gt; [ %{ &quot;text&quot; =&gt; &quot;This is a comment&quot; } , %{ &quot;text&quot; =&gt; &quot;This is another comment&quot; } ] , &quot;draft&quot; =&gt; %{ &quot;content&quot; =&gt; &quot;This is a draft blog post&quot; } , &quot;date&quot; =&gt; &quot;2021-11-11&quot; , &quot;time&quot; =&gt; &quot;14:00:00&quot; , &quot;metadata&quot; =&gt; %{ &quot;rating&quot; =&gt; 0 } } And now let's assume the struct we wish to make is this one: % BlogPost { content : &quot;This is a blog post&quot; , comments : [ % Comment { text : &quot;This is a comment&quot; } , % Comment { text : &quot;This is another comment&quot; } ] draft : % DraftPost { content : &quot;This is a draft blog post&quot; } } post_datetime : ~N[2020-11-11 14:00:00] } We can describe the following schemas to enable this: defmodule DraftPost do import DataSchema , only : [ data_schema : 1 ] data_schema ( [ field : { :content , &quot;content&quot; , &amp; to_string / 1 } ] ) end defmodule Comment do import DataSchema , only : [ data_schema : 1 ] data_schema ( [ field : { :text , &quot;text&quot; , &amp; to_string / 1 } ] ) def cast ( data ) do DataSchema . to_struct! ( data , __MODULE__ ) end end defmodule BlogPost do import DataSchema , only : [ data_schema : 1 ] data_schema ( [ field : { :content , &quot;content&quot; , &amp; to_string / 1 } , list_of : { :comments , &quot;comments&quot; , Comment } , has_one : { :draft , &quot;draft&quot; , DraftPost } , aggregate : { :post_datetime , %{ date : &quot;date&quot; , time : &quot;time&quot; } , &amp; BlogPost . to_datetime / 1 } , ] ) def to_datetime ( %{ date : date , time : time } ) do date = Date . from_iso8601! ( date ) time = Time . from_iso8601! ( time ) { :ok , datetime } = NaiveDateTime . new ( date , time ) datetime end end Then to transform some input data into the desired struct we can call DataSchema.to_struct!/2 which works recursively to transform the input data into the struct defined by the schema. source_data = %{ &quot;content&quot; =&gt; &quot;This is a blog post&quot; , &quot;comments&quot; =&gt; [ %{ &quot;text&quot; =&gt; &quot;This is a comment&quot; } , %{ &quot;text&quot; =&gt; &quot;This is another comment&quot; } ] , &quot;draft&quot; =&gt; %{ &quot;content&quot; =&gt; &quot;This is a draft blog post&quot; } , &quot;date&quot; =&gt; &quot;2021-11-11&quot; , &quot;time&quot; =&gt; &quot;14:00:00&quot; , &quot;metadata&quot; =&gt; %{ &quot;rating&quot; =&gt; 0 } } DataSchema . to_struct! ( source_data , BlogPost ) # This will output the following: % BlogPost { content : &quot;This is a blog post&quot; , comments : [ % Comment { text : &quot;This is a comment&quot; } , % Comment { text : &quot;This is another comment&quot; } ] draft : % DraftPost { content : &quot;This is a draft blog post&quot; } } post_datetime : ~N[2020-11-11 14:00:00] }","ref":"getting-started.html#creating-a-simple-schema","title":"What Are Data Schemas? - Creating A Simple Schema","type":"extras"},{"doc":"As we mentioned before we want to be able to handle multiple different kinds of source data in our schemas. For each type of source data we want to be able to specify how you access the data for each field type. We do that by providing a &quot;data accessor&quot; (a module that implements the DataSchema.DataAccessBehaviour ) when we create the schema. By default if you do not provide a specific data accessor module we use DataSchema.MapAccessor . That means the above example is equivalent to doing the following: defmodule DraftPost do import DataSchema , only : [ data_schema : 2 ] data_schema ( [ field : { :content , &quot;content&quot; , &amp; to_string / 1 } ] , DataSchema.MapAccessor ) end defmodule Comment do import DataSchema , only : [ data_schema : 2 ] data_schema ( [ field : { :text , &quot;text&quot; , &amp; to_string / 1 } ] , DataSchema.MapAccessor ) def cast ( data ) do DataSchema . to_struct! ( data , __MODULE__ ) end end defmodule BlogPost do import DataSchema , only : [ data_schema : 2 ] data_schema ( [ field : { :content , &quot;content&quot; , &amp; to_string / 1 } , list_of : { :comments , &quot;comments&quot; , Comment } , has_one : { :draft , &quot;draft&quot; , DraftPost } , aggregate : { :post_datetime , %{ date : &quot;date&quot; , time : &quot;time&quot; } , &amp; BlogPost . to_datetime / 1 } , ] , DataSchema.MapAccessor ) def to_datetime ( %{ date : date , time : time } ) do date = Date . from_iso8601! ( date ) time = Time . from_iso8601! ( time ) { :ok , datetime } = NaiveDateTime . new ( date , time ) datetime end end When creating the struct DataSchema will call the relevant function for the field we are creating, passing it the source data and the path to the value(s) in the source. Our DataSchema.MapAccessor looks like this: defmodule DataSchema.MapAccessor do @behaviour DataSchema.DataAccessBehaviour @impl true def field ( data , field ) do Map . get ( data , field ) end @impl true def list_of ( data , field ) do Map . get ( data , field ) end @impl true def has_one ( data , field ) do Map . get ( data , field ) end @impl true def aggregate ( data , field ) do Map . get ( data , field ) end end We can clean up our schema definitions a bit with currying. Instead of passing DataSchema.MapAccessor every time we create a schema we can define a helper function like so: defmodule DataSchema.Map do defmacro map_schema ( fields ) do quote do require DataSchema DataSchema . data_schema ( unquote ( fields ) , DataSchema.MapAccessor ) end end end Then change our schema definitions to look like this: defmodule DraftPost do import DataSchema.Map , only : [ map_schema : 1 ] map_schema ( [ field : { :content , &quot;content&quot; , &amp; to_string / 1 } ] ) end defmodule Comment do import DataSchema.Map , only : [ map_schema : 1 ] map_schema ( [ field : { :text , &quot;text&quot; , &amp; to_string / 1 } ] ) def cast ( data ) do DataSchema . to_struct! ( data , __MODULE__ ) end end defmodule BlogPost do import DataSchema.Map , only : [ map_schema : 1 ] map_schema ( [ field : { :content , &quot;content&quot; , &amp; to_string / 1 } , list_of : { :comments , &quot;comments&quot; , Comment } , has_one : { :draft , &quot;draft&quot; , DraftPost } , aggregate : { :post_datetime , %{ date : &quot;date&quot; , time : &quot;time&quot; } , &amp; BlogPost . to_datetime / 1 } , ] ) def to_datetime ( %{ date : date_string , time : time_string } ) do date = Date . from_iso8601! ( date_string ) time = Time . from_iso8601! ( time_string ) { :ok , datetime } = NaiveDateTime . new ( date , time ) datetime end end This means should we want to change how we access data (say we wanted to use Map.fetch! instead of Map.get ) we would only need to change the accessor used in one place - inside map_schema/1 .","ref":"getting-started.html#different-source-data-types","title":"What Are Data Schemas? - Different Source Data Types","type":"extras"},{"doc":"Transforming XML with XML schemas Let's imagine that we have some XML that we wish to turn into a struct. What would it require to enable that? First a new Xpath data accessor: defmodule XpathAccessor do @behaviour DataSchema.DataAccessBehaviour import SweetXml , only : [ sigil_x : 2 ] @impl true def field ( data , path ) do SweetXml . xpath ( data , ~x&quot; \#{ path } &quot;s ) end @impl true def list_of ( data , path ) do SweetXml . xpath ( data , ~x&quot; \#{ path } &quot;l ) end @impl true def has_one ( data , path ) do SweetXml . xpath ( data , ~x&quot; \#{ path } &quot; ) end @impl true def aggregate ( data , path ) do SweetXml . xpath ( data , ~x&quot; \#{ path } &quot;s ) end end defmodule DataSchema.Xpath do defmacro xpath_schema ( fields ) do quote do require DataSchema DataSchema . data_schema ( unquote ( fields ) , XpathAccessor ) end end end As we can see our accessor uses the library Sweet XML to access the XML. That means if we wanted to change the library later we would only need to alter this one module for all of our schemas to benefit from the change. Our source data looks like this: source_data = &quot;&quot;&quot; &lt;Blog date=&quot;2021-11-11&quot; time=&quot;14:00:00&quot;&gt; &lt;Content&gt;This is a blog post&lt;/Content&gt; &lt;Comments&gt; &lt;Comment&gt;This is a comment&lt;/Comment&gt; &lt;Comment&gt;This is another comment&lt;/Comment&gt; &lt;/Comments&gt; &lt;Draft&gt; &lt;Content&gt;This is a draft blog post&lt;/Content&gt; &lt;/Draft&gt; &lt;/Blog&gt; &quot;&quot;&quot; Let's define our schemas like so: defmodule DraftPost do import DataSchema.Xpath , only : [ xpath_schema : 1 ] xpath_schema ( [ field : { :content , &quot;./Content/text()&quot; , &amp; to_string / 1 } ] ) end defmodule Comment do import DataSchema.Xpath , only : [ xpath_schema : 1 ] xpath_schema ( [ field : { :text , &quot;./text()&quot; , &amp; to_string / 1 } ] ) def cast ( data ) do DataSchema . to_struct! ( data , __MODULE__ ) end end defmodule BlogPost do import DataSchema.Xpath , only : [ xpath_schema : 1 ] xpath_schema ( [ field : { :content , &quot;/Blog/Content/text()&quot; , &amp; to_string / 1 } , list_of : { :comments , &quot;//Comment&quot; , Comment } , has_one : { :draft , &quot;/Blog/Draft&quot; , DraftPost } , aggregate : { :post_datetime , %{ date : &quot;/Blog/@date&quot; , time : &quot;/Blog/@time&quot; } , &amp; BlogPost . to_datetime / 1 } , ] ) def to_datetime ( %{ date : date_string , time : time_string } ) do date = Date . from_iso8601! ( date_string ) time = Time . from_iso8601! ( time_string ) { :ok , datetime } = NaiveDateTime . new ( date , time ) datetime end end And now we can transform: source_data = &quot;&quot;&quot; &lt;Blog date=&quot;2021-11-11&quot; time=&quot;14:00:00&quot;&gt; &lt;Content&gt;This is a blog post&lt;/Content&gt; &lt;Comments&gt; &lt;Comment&gt;This is a comment&lt;/Comment&gt; &lt;Comment&gt;This is another comment&lt;/Comment&gt; &lt;/Comments&gt; &lt;Draft&gt; &lt;Content&gt;This is a draft blog post&lt;/Content&gt; &lt;/Draft&gt; &lt;/Blog&gt; &quot;&quot;&quot; DataSchema . to_struct! ( source_data , BlogPost ) # This will output: % BlogPost { comments : [ % Comment { text : &quot;This is a comment&quot; } , % Comment { text : &quot;This is another comment&quot; } ] , content : &quot;This is a blog post&quot; , draft : % DraftPost { content : &quot;This is a draft blog post&quot; } , post_datetime : ~N[2021-11-11 14:00:00] }","ref":"transforming-xml-with-xml-schemas.html","title":"Transforming XML with XML schemas","type":"extras"},{"doc":"There are two ways to create a struct from a schema DataSchema.to_struct!/2 and DataSchema.to_struct/2 . The crucial difference is that with the latter you can have your casting functions return an :error or an error tuple and the creation of the struct will halt and the error will be returned. In contrast DataSchema.to_struct! will always put whatever is returned from the casting functions into the struct so the only way to fail is to raise an error in the casting function. Contrast the following approaches: DataSchema.to_struct/2 defmodule BlagPost do import DataSchema , only : [ data_schema : 1 ] data_schema ( aggregate : { :post_datetime , %{ date : &quot;date&quot; , time : &quot;time&quot; } , &amp; BlagPost . to_datetime / 1 } ) def to_datetime ( %{ date : date_string , time : time_string } ) do with { :date , { :ok , date } } &lt;- { :date , Date . from_iso8601 ( date_string ) } , { :time , { :ok , time } } &lt;- { :time , Time . from_iso8601 ( time_string ) } , { :ok , datetime } &lt;- NaiveDateTime . new ( date , time ) do datetime else { :date , { :error , _ } } -&gt; { :error , &quot;Date is invalid: \#{ inspect ( date_string ) } &quot; } { :time , { :error , _ } } -&gt; { :error , &quot;Time is invalid: \#{ inspect ( time_string ) } &quot; } end end end input = %{ &quot;date&quot; =&gt; &quot;not a date&quot; , &quot;time&quot; =&gt; &quot;14:00:00&quot; , } DataSchema . to_struct ( input , BlagPost ) # =&gt; DataSchema.to_struct!/2 defmodule BlogPost do import DataSchema , only : [ data_schema : 1 ] data_schema ( aggregate : { :post_datetime , %{ date : &quot;date&quot; , time : &quot;time&quot; } , &amp; BlogPost . to_datetime / 1 } ) def to_datetime ( %{ date : date_string , time : time_string } ) do date = Date . from_iso8601! ( date_string ) time = Time . from_iso8601! ( time_string ) { :ok , datetime } = NaiveDateTime . new ( date , time ) datetime end end input = %{ &quot;date&quot; =&gt; &quot;not a date&quot; , &quot;time&quot; =&gt; &quot;14:00:00&quot; , } DataSchema . to_struct ( input , BlogPost ) # =&gt; ** (ArgumentError) cannot parse &quot;not a date&quot; as date, reason: :invalid_format","ref":"stopping-to_struct-when-there-is-an-error.html","title":"Stopping to_struct when there is an error","type":"extras"}]